#![allow(rustdoc::broken_intra_doc_links)]
//! Definitions for Ruby's special constants.
//!
//! Makes it easier to reference important Ruby constants, without havign to dig
//! around in bindgen's output.

pub use crate::ruby_value_type::*;
