//! Raw bindings to libruby, generated by bindgen.
//!
//! This module contains the raw bindings to libruby, generated by bindgen.

#![allow(non_upper_case_globals)]
#![allow(non_camel_case_types)]
#![allow(non_snake_case)]
#![allow(unknown_lints)]
#![allow(deref_nullptr)]
#![warn(unknown_lints)]
#![allow(clippy::all)]
#![allow(rustdoc::broken_intra_doc_links)]
#![allow(rustdoc::invalid_rust_codeblocks)]
#![allow(rustdoc::invalid_html_tags)]
#![allow(deprecated)]
#![allow(dead_code)]

include!(env!("RB_SYS_BINDINGS_PATH"));

pub use uncategorized::*;
pub use unstable::*;
